#include "debug.h"
#include "nand.h"
#include "in_nand.h"
#include "stat.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/stat.h>
#include <sys/types.h>

#define NAND_FREE_PAGE		-1

struct config_nand config_nand;
static struct blk_info *nandblk=NULL;

FILE *fp;

/*
 * find smallest bit number to represent number
 */
_u8 bit_find(_u32 number) {
	_u8 cnt=0;
	_u32 orig=number;
	while(number) {
		number>>=1;
		cnt++;
	}
	if(((_u32)(1<<(cnt-1)))<orig) return cnt;
	return cnt-1;
}

/*
 * initialization for NAND module
 * set NAND configuration and allocation data structure
 */
int nand_open(void) {
	unsigned int i,j;
	_u32 datablks;

	if(nandblk) nand_close();

	config_nand.sect_per_page = SECT_PER_PAGE;
    config_nand.page_per_blk = PAGE_PER_BLK;
	config_nand.totalsize = TOTALSIZE_GB;
	config_nand.extrablks = EXTRABLKS;
		
	datablks = config_nand.totalsize*1024*1024*2/config_nand.sect_per_page/config_nand.page_per_blk ;

	config_nand.numblks = datablks + config_nand.extrablks;

	config_nand.bit_sect_per_page = bit_find(config_nand.sect_per_page);
	config_nand.bit_page_per_blk = bit_find(config_nand.page_per_blk);

	nandblk = (struct blk_info *)malloc(sizeof(struct blk_info)*config_nand.numblks);
	if(nandblk == NULL) {
		error("Out of memory\n");
		return -1;
	}

	memset( nandblk, 0xFF, sizeof(struct blk_info) * config_nand.numblks);

	for(i=0;i<config_nand.numblks;i++) {
		nandblk[i].cnterase = 0;
		nandblk[i].numwrite = 0;
		nandblk[i].numvalid = 0;
		nandblk[i].last_touched = 0;
		nandblk[i].pages = (_u32 *)malloc(sizeof(_u32) * config_nand.page_per_blk);
		for(j=0;j<config_nand.page_per_blk;j++) {
			nandblk[i].pages[j]=NAND_FREE_PAGE;
		}
	};

	nand_stat_reset();
	
	return 0;
}

/*
 * deallocation data structure
 */
void nand_close() {
	unsigned int i;
	if(nandblk) {
		for(i=0;i<config_nand.numblks;i++) {
			if(nandblk[i].pages) free(nandblk[i].pages);
		}
		free(nandblk);
		nandblk = NULL;
	}

}

/*
 * count read operations
 *		blk_no: NAND block number to read
 *		page_no: NAND page number of a block to read
 *		lsn: logical sector number to verify mapping
 *           check if the lsn is the same value of the written lsn of page
 *		optStat: stat option to count (NAND_REQ_DATA or NAND_REQ_GC)
 */
_t_sect nand_page_read(unsigned int blk_no, unsigned int page_no,
    unsigned int lsn, int optStat) {

/* validation check: 
 *		- range of block number and page number
 *		- check logical sector number
 */
	assert(blk_no < config_nand.numblks);
	assert(page_no < config_nand.page_per_blk);
	if(optStat==NAND_REQ_DATA) {
		assert(nandblk[blk_no].pages[page_no] == lsn);
	}

	switch(optStat) {
	case NAND_REQ_GC:
		nand_stat.gc.read+=1;
		break;
	case NAND_REQ_DATA:
	default:
		nand_stat.data.read+=1;
		break;
	}

	return nandblk[blk_no].pages[page_no];
}

/*
 * count write operations
 *		blk_no: NAND block number to write
 *		page_no: NAND page number of a block to write
 *		lsn: logical sector number to verify mapping.
 *		optStat: stat option to count (NAND_REQ_DATA or NAND_REQ_GC)
 */
void nand_page_write(unsigned int blk_no, unsigned int page_no,
    unsigned int lsn, int optStat) {

/* validation check
 *		- range of block number and page number
 *		- ordered write for page number on a block
 *		- write on free page
 */
	assert(blk_no < config_nand.numblks);
	assert(page_no < config_nand.page_per_blk);
	assert(page_no >= nandblk[blk_no].last_touched);
	assert(nandblk[blk_no].pages[page_no] == NAND_FREE_PAGE);

	nandblk[blk_no].pages[page_no] = lsn;

	nandblk[blk_no].numwrite++;
	nandblk[blk_no].numvalid++;
	nandblk[blk_no].last_touched = page_no+1;

	switch(optStat) {
	case NAND_REQ_GC:
		nand_stat.gc.write+=1;
		break;
	case NAND_REQ_DATA:
	default:
		nand_stat.data.write+=1;
		break;
	}
}

/*
 * count erase operations
 *		blk_no: NAND block number to erase
 *		optStat: stat option to count (NAND_REQ_GC)
 */
void nand_block_erase(unsigned int blk_no, int optStat) {
	int i;

	assert(blk_no < config_nand.numblks);
	
	nandblk[blk_no].cnterase += 1;
	nandblk[blk_no].numwrite = 0;
	nandblk[blk_no].numvalid = 0;
	nandblk[blk_no].last_touched = 0;
	
	for(i=0;i<config_nand.page_per_blk;i++) {
		nandblk[blk_no].pages[i] = NAND_FREE_PAGE;
	}

	switch(optStat) {
	case NAND_REQ_DATA:
	case NAND_REQ_GC:
	default:
		nand_stat.gc.erase += 1;
		break;
	}
}


