#include <stdlib.h>
#include <string.h>
#include "types.h"
#include "debug.h"
#include "nand.h"
#include "vmbitmap.h"

/* 
 * bitmap implementation to store page invalidation information
 * used by garbage collection
 */

struct bitmap_t {
	int size;
	unsigned int *map;
} bitmap;

// one record can store 32 bit invalid information
#define BITS_PER_RECORD	32		
#define NUMSUB	((config_nand.page_per_blk + BITS_PER_RECORD - 1) / BITS_PER_RECORD)
#define CREATE_BITMAP_KEY(blkno, i)	((MUFTL_BITMAP_H) | (((blkno)*(NUMSUB)) + (i)))

/*
 * store page invalidation information for later garbage collection
 *		pbn: physical block number to invalidate
 *		ppn: physical page number of a block to invalidate
 *		length: the size of pages to invalidate
 */
void bitmap_s_invalidate_page(_t_blk pbn, _t_page ppn, _t_size length) {
	unsigned int i, start;
	int idx, offset;

	assert(length==1);

	start = (pbn*config_nand.page_per_blk + ppn);	

	for(i=start; i< start+length; i++) {
		idx = i / BITS_PER_RECORD;
		offset = i % BITS_PER_RECORD;

		assert((bitmap.map[idx] & (1<<offset)) == 0);	
		bitmap.map[idx] |= (1<<offset);	
	}

}

/*
 * helper function for bitmap_s_invalidate_page
 */
void bitmap_s_invalidate(_t_blk pbn, _t_page ppn, _t_size length) {

	bitmap_s_invalidate_page(pbn, ppn, length);
}

/* deallocate data structure */
void bitmap_s_end() {
	free(bitmap.map);
}

/* allocate data structure */
void bitmap_s_init() {

	bitmap.size = (config_nand.numblks * config_nand.page_per_blk + BITS_PER_RECORD -1) / BITS_PER_RECORD;
	bitmap.map = malloc(sizeof(int) * bitmap.size);

	memset(bitmap.map, 0, sizeof(int) * bitmap.size);	

	printf("Simple bitmap(using memory table) initialized\n");
}

/*
 * find bitmap information from physical block number for garbage collection
 *		pbn: physical block number of a request block
 *		block_bitmap: bitmap information of a request block(return)
 */
void bitmap_s_get_block(int pbn, _u8 *block_bitmap) {
	int i, bit;
	int start, idx, offset;

	start = (pbn*config_nand.page_per_blk);

	for(i=start; i<start+config_nand.page_per_blk;i++) {
		idx = i / BITS_PER_RECORD;
		offset = i % BITS_PER_RECORD;
		bit = (bitmap.map[idx] & (1<<offset)) >> offset;
		block_bitmap[i-start] = bit;
	}
}

/*
 * for erase block, reset bitmap information to reuse
 *		pbn: physical block number to erase
 */
void bitmap_s_clear_block(_t_blk pbn) {
	int i;
	int start, idx, offset;

	start = (pbn * config_nand.page_per_blk);

	for(i=start; i<start+config_nand.page_per_blk;i++) {
		idx = i / BITS_PER_RECORD;
		offset = i % BITS_PER_RECORD;

		bitmap.map[idx] &= (0<<offset);
	}

}
