#include "types.h"
#include "debug.h"
#include "nand.h"
#include "stat.h"
#include "vmbitmap.h"
#include "map.h"
#include "ftl.h"

/*
 * FTL read function
 *		lsn: logical sector number from trace file
 *		size: sector size from trace file
 *		return the size of sector read actually
 */
_t_size ftl_read(_t_sect lsn, _t_size size) {
	_t_size read_size;
	_t_sect read_lsn;
	_u32 lpa = LPA_F_SECT(lsn), ppa;

	ppa = map_find(lpa);
	assert(ppa!=-1);
	
	read_lsn = LSN_F_LPA(lpa);
	read_size = config_nand.sect_per_page - (lsn - read_lsn);

	nand_page_read(BLK_F_PA(ppa), PN_F_PA(ppa), read_lsn, NAND_REQ_DATA);

	return read_size;
}

/*
 * FTL write function
 *		lsn: logical sector number from trace file
 *		size: sector size from trace file
 *		return the size of sector written actually
 */
_t_size ftl_write(_t_sect lsn, _t_size size) {
	_t_size write_size;
	_t_sect write_lsn;
	_t_blk pbn;
	_t_page ppn;
	_u32 lpa = LPA_F_SECT(lsn), ppa;

/* find current written page(physical page) of request page(logical page) */
	ppa = map_find(lpa);
	assert(ppa!=-1);
	
	write_lsn = LSN_F_LPA(lpa);

/* check read-modify-write */
	if(ppa!=-1 && (write_lsn != lsn || size<config_nand.sect_per_page)) {
		nand_page_read(BLK_F_PA(ppa), PN_F_PA(ppa), write_lsn, NAND_REQ_DATA);
	}

/* allocate new page to write */
	pbn = (_t_blk)BLK_F_PA(ppa);
	vm_alloc_page(write_lsn, &pbn, &ppn, NAND_REQ_DATA);

	nand_page_write(pbn, ppn, write_lsn, NAND_REQ_DATA);

/* invalidate current page of request page */
	vm_invalidate(BLK_F_PA(ppa), PN_F_PA(ppa), 1);

/* update map */
	map_update(lpa, MAKE_PA(pbn, ppn));

/* check if garbage collection is needed */
	vm_gc_start();

	write_size = config_nand.sect_per_page - (lsn - write_lsn);
	
	return write_size;
}

/*
 * Warm up function
 * This FTL assumes that all data are written on flash before running trace.
 */
static void warm_all(_u32 blknum) {
	int write_size = blknum*config_nand.page_per_blk*config_nand.sect_per_page;
	int lsn = 0;
	_t_blk pbn;
	_t_page ppn;
	_u32 lpa;

	while(write_size) {
		lpa=LPA_F_SECT(lsn);
		vm_alloc_page(lsn, &pbn, &ppn, NAND_REQ_WARM);
		nand_page_write(pbn, ppn, lsn, NAND_REQ_WARM);
		map_update(lpa, MAKE_PA(pbn, ppn));

		lsn+=config_nand.sect_per_page;	
		write_size-=config_nand.sect_per_page;
	}
	
}

/*
 * FTL close function
 * deallocate data structure
 */
void ftl_close() {
	map_end();
	vm_end();
	bitmap_s_end();
}

/*
 * FTL open function
 * allocate data structure and start warm up
 */
int ftl_open() {

	bitmap_s_init();
	vm_init();
	
	map_init();

	warm_all(config_nand.numblks - config_nand.extrablks);

	printf("FTLSIMPLE initialized (total block: %d, extra: %d)\n", config_nand.numblks, config_nand.extrablks);
	printf("\n");

	return 0;

}

