#include <stdlib.h>
#include <string.h>
#include "debug.h"
#include "nand.h"
#include "vmbitmap.h"
#include "map.h"
#include "ftl.h"
#include "stat.h"

#define MIN_FREE_BLKS	10
#define GC_BLKS			5

int indexpages;

enum { BLOCK_FREE=0,
	BLOCK_DATA,
	BLOCK_INDEX,
	BLOCK_GC,
};

struct vbm_entry {
	_s32 invalid;
	_s32 map;
	_s32 score;
};

/*
 * current flash information for garbage collection and page allocation.
 * block information(FREE, DATA), invald bitmap information, 
 * and update block information
 */
static struct _t_vm {
	struct vbm_entry *vbm;

	_t_blk totblks;

	_u32 *update_blk_no;
	_u32 *update_page_no;
	unsigned int max_update_blk;
	int max_score;

	int fb_num;
	_t_blk cur_fb_no;
} vm;

/*
 * find new free block to write
 */
_t_blk vm_alloc_block(int reqtype) {
	unsigned int i;
	_t_blk blkno;
	if(vm.fb_num == 0) {
		error("No free blocks. Need GC?\n");
		assert(0);
	}

	for(i=0;i<vm.totblks; i++) {
		if(vm.vbm[vm.cur_fb_no].map==BLOCK_FREE) {
			blkno = vm.cur_fb_no;
			break;
		}
		vm.cur_fb_no++;
		if(vm.cur_fb_no>=vm.totblks) vm.cur_fb_no=0;
	}
	if(i==vm.totblks) {
		error("Can't find free block...\n");
		assert(0);
		return -1;
	}
	switch(reqtype) {
	case NAND_REQ_DATA:
	case NAND_REQ_GC:
	case NAND_REQ_WARM:
		vm.vbm[blkno].map=BLOCK_DATA;
		break;
	}
	vm.fb_num--;

	return blkno;
}

/*
 * find new page to write
 *		lsn: logical sector number to write
 *		pbn: found physical block number to write (return)
 *		ppn: found physical page number to write (return)
 *		reqtype: request type(NAND_REQ_DATA or NAND_REQ_GC)
 */
_u32 vm_alloc_page(_t_sect lsn, _t_blk *pbn, _t_page *ppn, int reqtype) {
	int i, idxupd;
	switch(reqtype) {
	case NAND_REQ_DATA:
	case NAND_REQ_GC:
		idxupd = 0;
		break;
	case NAND_REQ_WARM:
		idxupd = 0;
		break;
	default:
		idxupd = vm.max_update_blk-1;
		indexpages++;
		if(indexpages==34) {
			i=3;
		}
		break;
	}

	vm.update_page_no[idxupd]++;
/* if there is no pages to write on a current block, 
   find a new free block to write */
	if(vm.update_page_no[idxupd]>=config_nand.page_per_blk) {
		vm.update_blk_no[idxupd] = vm_alloc_block(reqtype);
		vm.update_page_no[idxupd] = 0;
	} 
	
	*pbn = vm.update_blk_no[idxupd];
	*ppn = vm.update_page_no[idxupd];

	assert(*pbn<config_nand.numblks);
	return 0;
}

/*
 * find victim block(has highest invalid pages without update block) 
 * to garbage collection
 */
_t_blk vm_get_victim() {
	unsigned int i, j;
	int maxinvalid=0;
	_t_blk	blkno;

	for(i=0;i<vm.totblks;i++) {
/* do not select free block for victim block */
		if(vm.vbm[i].map==BLOCK_FREE) continue;
/* select victim block of full page invalidation */
		if(vm.vbm[i].invalid>=config_nand.page_per_blk) {
			blkno=i;
			break;
		}

/* do not select update block for victim block */
		for(j=0;j<vm.max_update_blk;j++) {
			if(vm.update_blk_no[j]==i) {
				break;
			}
		}
		if(j!=vm.max_update_blk) continue;

/* find a block which has highest invalid pages */
		if(vm.vbm[i].invalid> maxinvalid) {
			maxinvalid = vm.vbm[i].invalid;
			blkno = i;
		}
	}

	return blkno;
}

/*
 * bitmap helper function
 * return sum of bitmap values of a block
 */
int bitmap_sum(_u8 *bitmap) {
	int i, res=0;
	for(i=0;i<config_nand.page_per_blk;i++) {
		if(bitmap[i]) res++;
	}
	return res;
}

/*
 * valid copy on victim block
 */
void vm_gc_run(_t_blk victim) {
	int i;
	_u32 opt;
	_u8 *block_bitmap;
	_t_sect lsn, lpa;
	_t_blk pbn;
	_t_page ppn;

	block_bitmap = (_u8 *)malloc(sizeof(_u8) * config_nand.page_per_blk);

/* get bitmap information of victim block */
	bitmap_s_get_block(victim, block_bitmap);
	assert(bitmap_sum(block_bitmap)==vm.vbm[victim].invalid);

	for(i=0; i< config_nand.page_per_blk; i++) {
		if(!block_bitmap[i]) {
/* if the page is valid page, execute valid copy */
			lsn = nand_page_read(victim, i, 0, NAND_REQ_GC);
			lpa = LPA_F_SECT(lsn);
			opt = 0;
			pbn = victim;
		
/* allocate new page to valid copy */	
			vm_alloc_page(lsn, &pbn, &ppn, NAND_REQ_GC);

			nand_page_write(pbn, ppn, lsn, NAND_REQ_GC);
			vm_invalidate(victim, i, 1);
			map_update(lpa, MAKE_PA(pbn, ppn));
		}
	}

	free(block_bitmap);
}

/*
 * erase a victim block to reuse
 */
void vm_clear_block(_t_blk victim, int type) {
	unsigned int i;
	assert(vm.vbm[victim].map!=BLOCK_FREE);
	bitmap_s_clear_block(victim);

	nand_block_erase(victim, type);
	vm.vbm[victim].map=BLOCK_FREE;
	vm.vbm[victim].invalid=0;
	vm.fb_num++;
	
	for(i=0;i<vm.max_update_blk;i++) {
		if(vm.update_blk_no[i]==victim) {
			vm.update_blk_no[i]=-1;
			vm.update_page_no[i] = config_nand.page_per_blk;
		}
	}
}

/*
 * invalidate page for garbage collection
 *		victim: victim block number
 *		ppn: victim page number of a block to invalidate
 *		length: the size of pages to invalidate
 */
void vm_invalidate(_t_blk victim, _t_page ppn, _t_size length) {
	vm.vbm[victim].invalid+=length;
	assert(vm.vbm[victim].invalid<=config_nand.page_per_blk);
	if(vm.vbm[victim].map==BLOCK_INDEX) {
		indexpages-=length;
	}

	bitmap_s_invalidate(victim, ppn, length);

/* early garbage collection if all pages of a block are invalidated */
	if(vm.vbm[victim].map==BLOCK_DATA && vm.vbm[victim].invalid==config_nand.page_per_blk) {
		vm_clear_block(victim, NAND_REQ_GC);
	}
}

int vm_gc_check() {
	return vm.fb_num < MIN_FREE_BLKS;
}

/*
 * check current state(if free block remains smaller than threshold)
 * and execute garbage collection
 */
void vm_gc_start() {
	_u8 *bitmap;
	_t_blk victim;

	if(!vm_gc_check()) {
		return;
	}
	
	bitmap = (_u8 *)malloc(sizeof(_u8)*config_nand.page_per_blk);
	while(vm.fb_num < MIN_FREE_BLKS+GC_BLKS) {
/* find victim block for garbage collection */
		victim = vm_get_victim();
		if(vm.vbm[victim].map==BLOCK_INDEX) {
			vm.vbm[victim].map=BLOCK_GC;
			bitmap_s_get_block(victim, bitmap);

			assert(bitmap_sum(bitmap)==vm.vbm[victim].invalid);
			indexpages-=config_nand.page_per_blk - vm.vbm[victim].invalid;
			map_gc(victim, bitmap);

			vm_clear_block(victim, NAND_REQ_DATA);
		} else {
			vm.vbm[victim].map=BLOCK_GC;
/* do valid copy */
			vm_gc_run(victim);
/* erase victim block */
			vm_clear_block(victim, NAND_REQ_DATA);
		}
	}

	free(bitmap);
}

/*
 * deallocate data structure
 */
void vm_end() {
	if(vm.update_blk_no) {
		free(vm.update_blk_no);
		vm.update_blk_no=NULL;
	}
	if(vm.update_page_no) {
		free(vm.update_page_no);
		vm.update_page_no=NULL;
	}
	if(vm.vbm) {
		free(vm.vbm);
		vm.vbm = NULL;
	}
}

/*
 * allocate data structure
 */
void vm_init() {
	unsigned int i;

	vm.max_update_blk = 2;
		
	vm.update_blk_no = (_u32 *)malloc(vm.max_update_blk * sizeof(_u32));
	vm.update_page_no = (_u32 *)malloc(vm.max_update_blk * sizeof(_u32));
	for(i=0;i<vm.max_update_blk;i++) {
		vm.update_blk_no[i] = -1;
		vm.update_page_no[i] = config_nand.page_per_blk;
	}

	vm.fb_num = config_nand.numblks;
	vm.cur_fb_no = 0;

	vm.totblks = config_nand.numblks;
	vm.vbm = (struct vbm_entry *)malloc(vm.totblks * sizeof(struct vbm_entry));
	for(i=0;i<vm.totblks;i++) {
		vm.vbm[i].invalid=0;
		vm.vbm[i].map = BLOCK_FREE;
		vm.vbm[i].score = 0;
	}
	
}
